#include <stdio.h>

struct node {
    struct node *root, *left, *right;
    int symbol, useflag, number;
};
typedef struct node Ttree[0x200];

static int buffer;
static char length;

static void setbit(char bit) {
    buffer <<= 1;
    buffer |= bit;
    length++;

    if (length == 8) {
        putchar(buffer);
        length = 0;
        buffer = 0;
    }
}

static void setchar(char field) {
    char i;
    for (i = 7; i >= 0; i--) setbit((field >> i) & 1);
}

static void setint(int field) {
    int i;
    for (i = 0x1f; i >= 0; i--) setbit((field >> i) & 1);
}

static char getbit(void) {
    char val;

    if (length == 0) {
        buffer = getchar();
        length = 8;
    }

    val = (buffer >> (length - 1)) & 1;
    length--;

    return val;
}

static int getint(int size) {
    int i, value;

    value = 0;
    for (i = 0; i < size; i++) {
        int bit = getbit();

        value <<= 1;
        value |= bit;
    }

    return value;
}

static int  min(Ttree tree, int n) {
    int i, min;

    for (min = 0; tree[min].useflag && min < n; min++)
        ;
    for (i = 0; i < n; i++)
        if (tree[i].useflag == 0 && tree[i].number < tree[min].number)
            min = i;

    return min;
}

static void create_tree(Ttree tree, int *k) {
    int border, i;

    border = *k - 1;
    for (i = 0; i < border; i++) {
        int m1, m2;

        m1 = min(tree, *k);
        tree[m1].useflag = 1;

        m2 = min(tree, *k);
        tree[m2].useflag = 1;

        tree[*k].number = tree[m1].number + tree[m2].number;
        tree[*k].symbol = 0x200;
        tree[*k].left = &tree[m1];
        tree[*k].right = &tree[m2];

        tree[m1].root = tree[m2].root = &tree[*k];

        (*k)++;
    }
}

static void restore_tree(Ttree tree, int *k, struct node *curr) {
    int bit;

    bit = getbit();
    if (buffer == EOF) return;

    if (bit == 1) {
        curr->symbol = 0x200;
        curr->number = 1;

        curr->left = &tree[++(*k)];
        tree[*k].root = curr;
        restore_tree(tree, k, curr->left);

        curr->right = &tree[++(*k)];
        tree[*k].root = curr;
        restore_tree(tree, k, curr->right);
    }
    else {
        int symbol = getint(8);
        curr->symbol = symbol;
        curr->number = 1;
    }
}

void prep2decomp(Ttree tree, int *n) {
    restore_tree(tree, n, &tree[*n]);
}

void prep2comp(Ttree tree, struct node *pointers[], int *n) {
    int i, a[0x100] = {0}, j;

    while ((i = getchar()) != EOF) a[i]++;
    
    j = 0;
    for (i = 0; i < 0x100; i++)
        if (a[i]) {
            tree[j].number = a[i];
            tree[j].symbol = i;
            pointers[i] = &tree[j++];
            (*n)++;
        }

    create_tree(tree, n);
}

static void print_tree(Ttree tree, struct node *root, unsigned *size) {
    static int counter;
    counter++;

    if (root == NULL) return;
    if (root->symbol != 0x200) {
        setbit(0);
        setchar(root->symbol);
        *size += root->number * (counter - 1);
    }
    else {
        setbit(1);
        print_tree(tree, root->left, size);
        print_tree(tree, root->right, size);
    }
    counter--;
}

void compress(Ttree tree, struct node *pointers[], struct node *root) {
    int symbol;
    unsigned size = 0;

    print_tree(tree, root, &size);
    if (size == 0) {
        setint(root->number);
        return;
    }
    setint(size);

    while ((symbol = getchar()) != EOF) {
        struct node *t = pointers[symbol];
        char code[0x100] = {0};
        int i = 0;

        while (t != root) {
            if (t->root->left == t) code[i++] = 0;
            else code[i++] = 1;
            t = t->root;
        }
        for (i--; i >= 0; i--) setbit(code[i]);
    }
    while (length != 0) setbit(0);
}

void decompress(Ttree tree, int k) {
    unsigned size;
    struct node *curr;

    size = getint(0x20);
    if (k == 0) {
        int i;
        for (i = 0; i < size - 1; i++) putchar(tree[0].symbol);
    }

    curr = &tree[0];
    while (buffer != EOF && size) {
        int bit;

        size--;
        bit = getbit();
        if (buffer == EOF) break;

        if (bit == 0) curr = curr->left;
        else curr = curr->right;

        if (curr->symbol != 0x200) {
            putchar(curr->symbol);
            curr = &tree[0];
        }
    }
}
